// ====================================================================
// 
//   ECM: The Extensible Container Model
//   Copyright (C) 2004 THALES
//   Contact: openccm-ecm@objectweb.org
//
//   This library is free software; you can redistribute it and/or
//   modify it under the terms of the GNU Lesser General Public
//   License as published by the Free Software Foundation; either
//   version 2.1 of the License, or any later version.
//
//   This library is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//   Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public
//   License along with this library; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
//   USA
//
//   Initial developer(s): Mathieu Vadet.
//   Initial Funding: IST COACH European project (IST-2001-34445)
//                    http://www.ist-coach.org 
//
// ====================================================================



package org.objectweb.ccm.runtime;

import org.objectweb.corba.runtime.*;

/**
 **  <p>Default DOM parser for the CORBA Component Descriptor (CCD) XML DTD.</p>
 **/
public class CCDParser
{
    // ccd 
    static final private String _class_name = "CCDParser";
    private org.w3c.dom.Document _doc;
    private java.util.HashMap _repid2sname;
    private java.util.HashMap _sid2uuid;
    private ExtComponentInstallationImpl _extci;

    // public fields set during the parsing
    // and accessible from the outside
    public String home_repid;
    public String component_repid;
    public java.util.HashMap sname2policies;

    // default constructor
    public
    CCDParser(org.w3c.dom.Document doc,
              ExtComponentInstallationImpl extci)
    {
        // ccd 
        _doc = doc;
        _repid2sname = new java.util.HashMap();
        _sid2uuid = new java.util.HashMap();
        _extci = extci;

        // public fields
        home_repid = null;
        component_repid = null;
        sname2policies = new java.util.HashMap();
    }

    //
    //  internal operations
    //

    static private org.w3c.dom.Document
    openAsDocument(String fname, String dtdloc)
    {
        final String opname = "openAsDocument";
        // create SAX input source
        org.xml.sax.InputSource isource = null;
        try {
            isource = new org.xml.sax.InputSource(new java.io.FileInputStream(fname));
        } catch (java.io.FileNotFoundException ex) {
            final String msg = "FAILED (file not found: "+fname+")";
            TheLogger.error(_class_name, opname, msg, ex);
            return null;
        }

        // change location of DTD in document
        String loc = isource.getSystemId();
        isource.setSystemId(dtdloc+"/"+loc);

        // obtain new DOM parser factory
        // NOTE: disable namespaces and validation
        javax.xml.parsers.DocumentBuilderFactory fact = javax.xml.parsers.DocumentBuilderFactory.newInstance();
        fact.setNamespaceAware(false);
        fact.setValidating(false);

        // create DOM parser and parse XML file
        try {
            javax.xml.parsers.DocumentBuilder parser = fact.newDocumentBuilder();
            return parser.parse(isource);
        } catch (Exception ex) {
            TheLogger.error(_class_name, opname, "FAILED", ex);
            return null;
        }
    }    

    final private void
    parseDocument()
    {
        org.w3c.dom.Element root = _doc.getDocumentElement();
        org.w3c.dom.NodeList childs = root.getChildNodes();
        org.w3c.dom.Node node = null;
        for (int i=0;i<childs.getLength();i++) {
            node = childs.item(i);
            switch (node.getNodeType()) {
            case org.w3c.dom.Node.ELEMENT_NODE: 
                org.w3c.dom.Element element = (org.w3c.dom.Element)node;
                String tagname = element.getTagName();

                if (tagname.equals("usedservices")) {
                    // equiv XPath: "corbacomponent/usedservices"
                    parseUsedServices(element);
                } else if (tagname.equals("policies")) {
                    // equiv XPath: "corbacomponent/policies"
                    parsePolicies(element, "");
                } else if (tagname.equals("homefeatures")) {
                    // equiv XPath: "corbacomponent/homefeatures"
                    parseHome(element);
                } else if (tagname.equals("componentfeatures")) {
                    // equiv XPath: "corbacomponent/componentfeatures"
                    parseComponent(element);
                } else if (tagname.equals("interface")) {
                    // equiv XPath: "corbacomponent/interface"
                    parseInterface(element);
                } else if (tagname.equals("componentrepid")) {
                    // equiv XPath: "corbacomponent/componentrepid"
                    component_repid = element.getAttribute("repid");
                } else if (tagname.equals("homerepid")) {
                    // equiv XPath: "corbacomponent/homerepid"
                    home_repid = element.getAttribute("repid");
                }

                break;
            default:
                break;
            }
        }
    }

    final private void
    parseUsedServices(org.w3c.dom.Element uservices)
    {
        org.w3c.dom.NodeList services = uservices.getElementsByTagName("usedservice");
        org.w3c.dom.Element service = null;
        org.w3c.dom.Element implref = null;
        String serviceid = null;
        String uuid = null;

        for (int i=0;i<services.getLength();i++) {
            // equiv XPath: "corbacomponent/usedservices/usedservice[@id=X]"
            service = (org.w3c.dom.Element)services.item(i);
            // equiv XPath: "corbacomponent/usedservices/usedservice[@id=X]/@id"
            serviceid = service.getAttribute("id");
            // equiv XPath: "corbacomponent/usedservices/usedservice[@id=X]/implementationref"
            implref = (org.w3c.dom.Element)service.getElementsByTagName("implementationref").item(0);
            // equiv XPath: "corbacomponent/usedservices/usedservice[@id=X]/implementationref/@idref"
            uuid = implref.getAttribute("idref");

            // store
            _sid2uuid.put(serviceid, uuid);
        }
    }

    final private void
    parsePolicies(org.w3c.dom.Element pols, String sname)
    {
        final String opname = "parsePolicies";
        if (pols==null) {
            final String msg = "no policies element for target: "+sname;
            TheLogger.debug(_class_name, opname, msg);
            return ;
        }

        TheLogger.debug(_class_name, opname, "scoped name is: "+sname);

        org.w3c.dom.NodeList targets = pols.getElementsByTagName("target");
        org.w3c.dom.Element target = null;
        org.w3c.dom.NodeList servicerefs = null;
        org.w3c.dom.Element serviceref = null;
        String targetname = null;

        for (int i=0;i<targets.getLength();i++) {
            target = (org.w3c.dom.Element)targets.item(i);
            // equiv XPath: "policies/target/@name"
            targetname = target.getAttribute("name");

            // equiv XPath: "policies/target/serviceref"
            servicerefs = target.getElementsByTagName("serviceref");

            org.coach.ECM.CCDPolicyRef[] prefs = new org.coach.ECM.CCDPolicyRef[servicerefs.getLength()];
            org.coach.ECM.ServiceArchive sarchive = null;
            String idref = null;
            String policyref = null;
            String policyvalue = null;
            String uuid = null;

            for (int j=0;j<servicerefs.getLength();j++) {
                serviceref = (org.w3c.dom.Element)servicerefs.item(j);
                // equiv XPath: "policies/target/serviceref[@idref=X]/@idref"
                idref = serviceref.getAttribute("idref");
                // equiv XPath: "policies/target/serviceref[@idref=X]/@policyref"
                policyref = serviceref.getAttribute("policyref");
                // equiv XPath: "policies/target/serviceref[@idref=X]/@policyvalue"
                policyvalue = serviceref.getAttribute("policyvalue");

                uuid = (String)_sid2uuid.get(idref);
                sarchive = _extci.get_service_archive(uuid);

                //String msg = "policy found";
                //TheLogger.debug(_class_name, opname, msg);
                //msg="     target:     "+targetname;
                //TheLogger.debug(_class_name, opname, msg);
                //msg="     idref:      "+idref;
                //TheLogger.debug(_class_name, opname, msg);
                //msg="     policyref:  "+policyref;
                //TheLogger.debug(_class_name, opname, msg);
                //msg="     policyvalue:"+policyvalue;
                //TheLogger.debug(_class_name, opname, msg);

                prefs[j] = new org.coach.ECM.CCDPolicyRef();
                prefs[j].implementationref = uuid;
                prefs[j].service_id = sarchive.service_id();
                prefs[j].policyref = policyref;
                prefs[j].policyvalue = policyvalue;
            }

            // store
            sname2policies.put(sname+"::"+targetname, prefs);
        }
    }

    final private void
    parseEventPolicies(org.w3c.dom.Element pols, String sname)
    {
        // NOTE: for event ports, the <serviceref> element is a child of the
        // <policies> element

        final String opname = "parseEventPolicies";
        if (pols==null) {
            final String msg = "no policies element for target: "+sname;
            TheLogger.debug(_class_name, opname, msg);
            return ;
        }

        TheLogger.debug(_class_name, opname, "scoped name is: "+sname);

        org.w3c.dom.NodeList servicerefs = null;
        org.w3c.dom.Element serviceref = null;

        // equiv XPath: "policies/serviceref"
        servicerefs = pols.getElementsByTagName("serviceref");

        org.coach.ECM.CCDPolicyRef[] prefs = new org.coach.ECM.CCDPolicyRef[servicerefs.getLength()];
        org.coach.ECM.ServiceArchive sarchive = null;
        String idref = null;
        String policyref = null;
        String policyvalue = null;
        String uuid = null;

        for (int j=0;j<servicerefs.getLength();j++) {
            serviceref = (org.w3c.dom.Element)servicerefs.item(j);
            // equiv XPath: "policies/serviceref[@idref=X]/@idref"
            idref = serviceref.getAttribute("idref");
            // equiv XPath: "policies/serviceref[@idref=X]/@policyref"
            policyref = serviceref.getAttribute("policyref");
            // equiv XPath: "policies/serviceref[@idref=X]/@policyvalue"
            policyvalue = serviceref.getAttribute("policyvalue");

            uuid = (String)_sid2uuid.get(idref);
            sarchive = _extci.get_service_archive(uuid);

            //String msg = "policy found";
            //TheLogger.debug(_class_name, opname, msg);
            //msg="     idref:      "+idref;
            //TheLogger.debug(_class_name, opname, msg);
            //msg="     policyref:  "+policyref;
            //TheLogger.debug(_class_name, opname, msg);
            //msg="     policyvalue:"+policyvalue;
            //TheLogger.debug(_class_name, opname, msg);

            prefs[j] = new org.coach.ECM.CCDPolicyRef();
            prefs[j].implementationref = uuid;
            prefs[j].service_id = sarchive.service_id();
            prefs[j].policyref = policyref;
            prefs[j].policyvalue = policyvalue;
        }

        // store
        sname2policies.put(sname, prefs);
    }

    final private void
    parseHome(org.w3c.dom.Element home)
    {
        // equiv XPath: "corbacomponent/homefeatures[@name=X]/@name"
        String name = home.getAttribute("name");
        // equiv XPath: "corbacomponent/homefeatures[@name=X]/@repid"
        String repid = home.getAttribute("repid");
        // store
        _repid2sname.put(repid, "::"+name);

        // check home attributes/operations/finders/factories
        // equiv XPath: "corbacomponent/homefeatures[@name=X]/policies"
        org.w3c.dom.Element pols = (org.w3c.dom.Element)home.getElementsByTagName("policies").item(0);
        parsePolicies(pols, "::"+name);

        // check inherited home
        // TODO
    }

    final private void
    parseComponent(org.w3c.dom.Element component)
    {
        // equiv XPath: "corbacomponent/componentfeatures[@name=X]/@name"
        String name = component.getAttribute("name");
        // equiv XPath: "corbacomponent/componentfeatures[@name=X]/@repid"
        String repid = component.getAttribute("repid");
        // store
        _repid2sname.put(repid, "::"+name);

        // check component attributes/operations/ports
        // equiv XPath: "corbacomponent/componentfeatures[@name=X]/policies"
        // NOTE: CAUTION: the "getElementsByTagName" operation returns all the children, including grand-children
        // so if we call "getElementsByTagName" directly on the component element, we'll obtain the policies elements
        // defined in the ports element also
        // Therefore we use the "getChildNodes" operation instead
        org.w3c.dom.NodeList childs = component.getChildNodes();
        for (int i=0;i<childs.getLength();i++) {
            if (childs.item(i).getNodeName().equals("policies")) {
                org.w3c.dom.Element pols = (org.w3c.dom.Element)childs.item(i);
                parsePolicies(pols, "::"+name);
            }
        }

        // check inherited component
        // TODO

        // check supported interfaces
        // TODO

        // check ports
        // check receptacles
        parsePort(component, "::"+name, "uses", "usesname");
        // check facets
        parsePort(component, "::"+name, "provides", "providesname");
        // check event sources
        parseEventPort(component, "::"+name, "emits", "emitsname");
        parseEventPort(component, "::"+name, "publishes", "publishesname");
        // check event sink
        parseEventPort(component, "::"+name, "consumes", "consumesname");
    }

    final private void
    parsePort(org.w3c.dom.Element component, String cname,
              String ptype, String pattr)
    {
        org.w3c.dom.NodeList ports = null;
        org.w3c.dom.Element port = null;
        org.w3c.dom.Element pols = null;
        String pname = null;
        String prepid = null;
        // equiv XPath: "corbacomponent/componentfeatures[@name=X]/ports/<ptype>"
        ports = component.getElementsByTagName(ptype);
        for (int i=0;i<ports.getLength();i++) {
            port = (org.w3c.dom.Element)ports.item(i);
            // equiv XPath: "corbacomponent/componentfeatures[@name=X]/ports/<ptype>[@<pattr>=X]/@<pattr>"
            pname = port.getAttribute(pattr);
            // equiv XPath: "corbacomponent/componentfeatures[@name=X]/ports/<ptype>[@<pattr>=X]/@repid"
            prepid = port.getAttribute("repid");
            // check if interface pointed by "repid" has policies
            // TODO

            // check attributes/operations on port
            pols = (org.w3c.dom.Element)port.getElementsByTagName("policies").item(0);
            parsePolicies(pols, cname+"::"+pname);
        }
    }

    final private void
    parseEventPort(org.w3c.dom.Element component, String cname,
                   String ptype, String pattr)
    {
        org.w3c.dom.NodeList ports = null;
        org.w3c.dom.Element port = null;
        org.w3c.dom.Element pols = null;
        String pname = null;
        String pevt = null;
        // equiv XPath: "corbacomponent/componentfeatures[@name=X]/ports/<ptype>"
        ports = component.getElementsByTagName(ptype);
        for (int i=0;i<ports.getLength();i++) {
            port = (org.w3c.dom.Element)ports.item(i);
            // equiv XPath: "corbacomponent/componentfeatures[@name=X]/ports/<ptype>[@<pattr>=X]/@<pattr>"
            pname = port.getAttribute(pattr);
            // equiv XPath: "corbacomponent/componentfeatures[@name=X]/ports/<ptype>[@<pattr>=X]/@repid"
            pevt = port.getAttribute("eventtype");
            // NOTE: TODO: is there a XML element that describe the eventtype ?
            // if yes, than we should check this element for policies

            // check policies on port
            // NOTE: for event ports, policies are directly set on the port (and not on the operation
            // of the equivalent interface)
            pols = (org.w3c.dom.Element)port.getElementsByTagName("policies").item(0);
            parseEventPolicies(pols, cname+"::"+pname);
        }
    }

    final private void
    parseInterface(org.w3c.dom.Element itf)
    {
        // TODO
    }

    //
    //  static public operations
    //

    static public CCDParser
    parse(String fname, String dtdloc, ExtComponentInstallationImpl extci)
    {
        CCDParser parser = null;
        try {
            org.w3c.dom.Document doc = openAsDocument(fname, dtdloc);
            parser = new CCDParser(doc, extci);
            parser.parseDocument();
        } catch (Exception ex) {
            final String opname = "parse";
            TheLogger.error(_class_name, opname, "FAILED", ex);
            return null;
        }

        return parser;
    }
}
