// ====================================================================
// 
//   ECM: The Extensible Container Model
//   Copyright (C) 2004 THALES
//   Contact: openccm-ecm@objectweb.org
//
//   This library is free software; you can redistribute it and/or
//   modify it under the terms of the GNU Lesser General Public
//   License as published by the Free Software Foundation; either
//   version 2.1 of the License, or any later version.
//
//   This library is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//   Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public
//   License along with this library; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
//   USA
//
//   Initial developer(s): Mathieu Vadet.
//   Initial Funding: IST COACH European project (IST-2001-34445)
//                    http://www.ist-coach.org 
//
// ====================================================================



package org.objectweb.ccm.runtime;

import org.objectweb.corba.runtime.*;
// import of OpenCCM modules
import org.objectweb.corba.ast.api.*;
import org.objectweb.corba.ast.lib.*;
import org.objectweb.corba.generator.java.lib.*;
import org.objectweb.corba.generator.java.ast.api.*;
import org.objectweb.corba.generator.java.ast.lib.*;
import org.objectweb.corba.generator.common.lib.GenerationException;

/**
 **  <p>Generator class for home and component implementations.</p>
 **/
public class CIDLJImplGenerator
extends org.objectweb.corba.generator.common.lib.GeneratorBase
{
    // generator
    private ExtendedJavaTranslator _translator;
    private RepositoryImpl _repository; 

    // default constructor
    public
    CIDLJImplGenerator(AST ast)
    {
        super(ast);

        _translator = new ExtendedJavaTranslator();
        _repository = null;
    }

    //
    //  internal operations
    //

    private ExtendedJavaTranslator
    getTranslator()
    {
        return _translator;
    }

    private RepositoryImpl
    getRepository()
    {
        return _repository;
    }

    private void
    generateForDeclaration(Declaration decl)
    throws GenerationException
    {
        // generation is done only for: 
        //  - module
        //  - component
        //  - home
        //  - composition
        if (decl.getDeclKind()==DeclarationKind.dk_module) {
            generateForModule((ModuleDecl)decl);
        } else if (decl.getDeclKind()==DeclarationKind.dk_component) {
            generateForComponent((ComponentDecl)decl);
        } else if (decl.getDeclKind()==DeclarationKind.dk_home) {
            generateForHome((HomeDecl)decl);
        } else if (decl.getDeclKind()==DeclarationKind.dk_composition) {
            generateForComposition((CompositionDecl)decl);
        } 
    }

    private void
    generateForModule(ModuleDecl module)
    throws GenerationException
    {
        Declaration[] decls = module.getContents(false, DeclarationKind.dk_all);
        for (int i=0;i<decls.length;i++) {
            generateForDeclaration(decls[i]);
        }
    }

    private void
    generateForComposition(CompositionDecl composition)
    throws GenerationException
    {
        // TODO
    }

    private void
    generateForHome(HomeDecl home)
    throws GenerationException
    {
        String classname = home.getName()+"_Base";
        ClassObject ski = new ClassObjectImpl(classname);

        // header
        ski.setPackage(getTranslator().getPackage(home));
        ski.addImport("org.objectweb.corba.runtime.*");
        ski.addImport("org.objectweb.ccm.runtime.*");

        // class decl
        // * public class $classname
        //      extends KeylessCCMHomeBase
        //      implements org.omg.Components.Servant, $home.getName()Operations
        ski.addComment(" ");
        ski.addComment("Java servant implementation for home "+home.getName()); 
        ski.addComment(" ");
        ski.addComment("NOTE: generated class; DO NOT EDIT !");
        ski.addComment(" ");
        ski.setAbstract(true);
        ski.setModifier(ModifierKindImpl.mk_public);
        ski.addInheritedObject("KeylessCCMHomeBase");
        ski.addImplementedObject("org.omg.Components.Servant");
        ski.addImplementedObject(home.getName()+"Operations");

        // internal state
        AttributeObject attr = null;
        // * static private _class_name = "<class_name>";
        attr = new AttributeObjectImpl();
        attr.addComment("Class name of this class (used for debugging).");
        attr.setStatic(true);
        attr.setModifier(ModifierKindImpl.mk_private);
        attr.setType("String");
        attr.setName("_class_name");
        attr.setInitialValue("\""+classname+"\"");
        ski.addAttribute(attr);

        // default constructor
        // * public $classname() { ... }
        ConstructorObject constructor = new ConstructorObjectImpl();
        // initialize all fields
        constructor.getImpl().setMacro("EMPTY_BODY");
        ski.addConstructor(constructor);

        // abstract operations implementation
        // * final protected String
        //   getComponentRepid()
        MethodObject operation = null;
        operation = new MethodObjectImpl();
        operation.setFinal(true);
        operation.setModifier(ModifierKindImpl.mk_protected);
        operation.setReturnType("String");
        operation.setName("getComponentRepid");
        operation.getImpl().setMacro("HOMEBASE_GETCOMPREPID_BODY");
        operation.getImpl().addContextValue("component", home.getManagedComponent());
        ski.addMethod(operation);

        // public operations
        // * final public org.omg.PortableServer.Servant as_native_servant()
        operation = new MethodObjectImpl();
        operation.setFinal(true);
        operation.setModifier(ModifierKindImpl.mk_public);
        operation.setReturnType("org.omg.PortableServer.Servant");
        operation.setName("as_native_servant");
        operation.getImpl().setMacro("HOMEBASE_ASNATIVESERV_BODY");
        operation.getImpl().addContextValue("home", home);
        ski.addMethod(operation);

        // home implicit operations
        // * final public Client create()
        //   throws org.omg.Components.CreateFailure
        operation = new MethodObjectImpl();
        operation.setFinal(true);
        operation.setModifier(ModifierKindImpl.mk_public);
        operation.setReturnType(home.getManagedComponent().getName());
        operation.setName("create");
        operation.getImpl().setMacro("HOMEBASE_CREATE_BODY");
        operation.getImpl().addContextValue("component", home.getManagedComponent());
        operation.addException("org.omg.Components.CreateFailure");
        ski.addMethod(operation);

        // home explicit operations (factory, finder, operations)
        // NOTE: delegated to home implementation by inheritance

        // add to repository
        getRepository().addObject(ski);
    }

    private void
    generateForComponent(ComponentDecl comp)
    throws GenerationException
    {
        // NOTE: currently, we assume monolithic implementation (single segment)

        String classname = comp.getName()+"_MainSegBase";
        ClassObject ski = new ClassObjectImpl(classname);

        // header
        ski.setPackage(getTranslator().getPackage(comp));
        ski.addImport("org.objectweb.corba.runtime.*");
        ski.addImport("org.objectweb.ccm.runtime.*");

        // class decl
        // * public class $classname
        //      extends CCMObjectBase
        //      implements org.omg.Components.Servant, $component.getName()_MainSegOperations
        ski.addComment(" ");
        ski.addComment("Java servant implementation for component "+comp.getName()); 
        ski.addComment(" ");
        ski.addComment("NOTE: generated class; DO NOT EDIT !");
        ski.addComment(" ");
        ski.setAbstract(true);
        ski.setModifier(ModifierKindImpl.mk_public);
        ski.addInheritedObject("CCMObjectBase");
        ski.addImplementedObject("org.omg.Components.Servant");
        ski.addImplementedObject(comp.getName()+"_MainSegOperations");

        // internal state
        AttributeObject attr = null;
        // * static private _class_name = "<class_name>";
        attr = new AttributeObjectImpl();
        attr.addComment("Class name of this class (used for debugging).");
        attr.setStatic(true);
        attr.setModifier(ModifierKindImpl.mk_private);
        attr.setType("String");
        attr.setName("_class_name");
        attr.setInitialValue("\""+classname+"\"");
        ski.addAttribute(attr);

        // default constructor
        // * public $classname()
        ConstructorObject constructor = new ConstructorObjectImpl();
        // initialize all fields
        constructor.getImpl().setMacro("EMPTY_BODY");
        ski.addConstructor(constructor);

        // abstract operations implementation
        // * final protected void 
        //   initializePorts()
        MethodObject operation = new MethodObjectImpl();
        operation.setFinal(true);
        operation.setModifier(ModifierKindImpl.mk_protected);
        operation.setReturnType("void");
        operation.setName("initializePorts");
        operation.getImpl().setMacro("MAINSEGBASE_INITPORTS_BODY");
        Declaration[] recs = comp.getContents(false, DeclarationKind.dk_uses);
        Declaration[] facets = comp.getContents(false, DeclarationKind.dk_provides);
        Declaration[] sinks = comp.getContents(false, DeclarationKind.dk_consumes);
        Declaration[] emitters = comp.getContents(false, DeclarationKind.dk_emits);
        Declaration[] publishers = comp.getContents(false, DeclarationKind.dk_publishes);
        if (facets.length!=0) {
            operation.getImpl().addContextValue("has_facets", new Boolean(true));
            operation.getImpl().addContextValue("facets", facets);
            operation.getImpl().addContextValue("facetslength", Integer.toString(facets.length));
        }
        else {
            operation.getImpl().addContextValue("has_facets", new Boolean(false));
        }
        if (recs.length!=0) {
            operation.getImpl().addContextValue("has_recs", new Boolean(true));
            operation.getImpl().addContextValue("recs", recs);
            operation.getImpl().addContextValue("recslength", Integer.toString(recs.length));
        }
        else {
            operation.getImpl().addContextValue("has_recs", new Boolean(false));
        }
        if (sinks.length!=0) {
            operation.getImpl().addContextValue("has_sinks", new Boolean(true));
            operation.getImpl().addContextValue("sinks", sinks);
            operation.getImpl().addContextValue("sinkslength", Integer.toString(sinks.length));
        }
        else {
            operation.getImpl().addContextValue("has_sinks", new Boolean(false));
        }
        if (emitters.length!=0) {
            operation.getImpl().addContextValue("has_emitters", new Boolean(true));
            operation.getImpl().addContextValue("emitters", emitters);
            operation.getImpl().addContextValue("emitterslength", Integer.toString(emitters.length));
        }
        else {
            operation.getImpl().addContextValue("has_emitters", new Boolean(false));
        }
        if (publishers.length!=0) {
            operation.getImpl().addContextValue("has_publishers", new Boolean(true));
            operation.getImpl().addContextValue("publishers", publishers);
            operation.getImpl().addContextValue("publisherslength", Integer.toString(publishers.length));
        }
        else {
            operation.getImpl().addContextValue("has_publishers", new Boolean(false));
        }

        ski.addMethod(operation);

        // * final public org.omg.PortableServer.Servant as_native_servant()
        operation = new MethodObjectImpl();
        operation.setFinal(true);
        operation.setModifier(ModifierKindImpl.mk_public);
        operation.setReturnType("org.omg.PortableServer.Servant");
        operation.setName("as_native_servant");
        operation.getImpl().setMacro("MAINSEGBASE_ASNATIVESERV_BODY");
        operation.getImpl().addContextValue("component", comp);
        ski.addMethod(operation);

        // facets equiv operation
        // * final public ${facet.getInterface().getName()} provide_${facet.getName()}()
        for (int i=0;i<facets.length;i++) {
            // find the operation declarations
            // normally, there's only one declaration
            ProvidesDecl facet = (ProvidesDecl)facets[i];
            OperationDecl opprovide = facet.getClientMapping()[0];

            addJavaOperation(ski, opprovide, opprovide.getType(),
                             opprovide.getParameterList().getParameters(),
                             opprovide.getExceptionList().getExceptions(),
                             "MAINSEGBASE_PROVIDEFACET_OPERATION_BODY", "facet", facet);
        }

        // receptacles equiv operation
        for (int i=0;i<recs.length;i++) {
            UsesDecl receptacle = (UsesDecl)recs[i];
            // check multiple; do not use the same body macro
            String macrobody = "SIMPLE";
            if (receptacle.isMultiple()) {
                macrobody = "MULTIPLE";
            }

            // find the operation declarations
            OperationDecl opconnect = null;
            OperationDecl opdisconnect = null;
            OperationDecl opgetconnection = null;
            OperationDecl[] opeqs = receptacle.getClientMapping();
            for (int j=0;j<opeqs.length;j++) {
                if (opeqs[j].getName().charAt(0)=='c') {
                    opconnect = opeqs[j];
                }
                else if (opeqs[j].getName().charAt(0)=='d') {
                    opdisconnect = opeqs[j];
                }
                else if (opeqs[j].getName().charAt(0)=='g') {
                    opgetconnection = opeqs[j];
                }
            }

            // * final public void
            //   connect_${rec.getName()}(${rec.getInterface().getName()} connexion)
            //   throws org.omg.Components.AlreadyConnected
            // or
            // * final public org.omg.Components.Cookie
            //   connect_${rec.getName()}(${rec.getInterface().getName()} connexion)
            //   throws org.omg.Components.ExceededConnectionLimit
            addJavaOperation(ski, opconnect, opconnect.getType(),
                             opconnect.getParameterList().getParameters(),
                             opconnect.getExceptionList().getExceptions(),
                             "MAINSEGBASE_"+macrobody+"_CONNECT_OPERATION_BODY",
                             "receptacle", receptacle);

            // * final public ${rec.getInterface().getName()}
            //   disconnect_${rec.getName()}()
            //   throws org.omg.Components.NoConnection
            // or
            // * final public ${rec.getInterface().getName()}
            //   disconnect_${rec.getName()}()
            //   throws org.omg.Components.InvalidConnection
            addJavaOperation(ski, opdisconnect, opdisconnect.getType(),
                             opdisconnect.getParameterList().getParameters(),
                             opdisconnect.getExceptionList().getExceptions(),
                             "MAINSEGBASE_"+macrobody+"_DISCONNECT_OPERATION_BODY",
                             "receptacle", receptacle);

            // * final public ${rec.getInterface().getName()}
            //   get_connection_${rec.getName()}()
            // or
            // * final public ${rec.getName()}Connections
            //   get_connections_${rec.getName()}()
            if (receptacle.isMultiple()) {
                String[] ctxkeys = { "receptacle", "comp_absname" };
                Object[] ctxvals = { receptacle, getTranslator().getAbsoluteName(comp) };

                addJavaOperation(ski, opgetconnection, opgetconnection.getType(),
                                 opgetconnection.getParameterList().getParameters(),
                                 opgetconnection.getExceptionList().getExceptions(),
                                 "MAINSEGBASE_"+macrobody+"_GETCONNECTION_OPERATION_BODY",
                                 ctxkeys, ctxvals);
            } else {
                addJavaOperation(ski, opgetconnection, opgetconnection.getType(),
                                 opgetconnection.getParameterList().getParameters(),
                                 opgetconnection.getExceptionList().getExceptions(),
                                 "MAINSEGBASE_"+macrobody+"_GETCONNECTION_OPERATION_BODY",
                                 "receptacle", receptacle);
            }
        }

        // sinks equiv operation
        // * final public ${sink.getEvent().getName()}Consumer get_consumer_${sink.getName()}()
        for (int i=0;i<sinks.length;i++) {
            // find the operation declarations
            // normally, there's only one declaration
            ConsumesDecl sink = (ConsumesDecl)sinks[i];
            OperationDecl opsink = sink.getClientMapping()[0];

            addJavaOperation(ski, opsink, opsink.getType(),
                             opsink.getParameterList().getParameters(),
                             opsink.getExceptionList().getExceptions(),
                             "MAINSEGBASE_GETCONSUMER_OPERATION_BODY", "sink", sink);
        }

        // emitters equiv operation
        for (int i=0;i<emitters.length;i++) {
            EmitsDecl emitter = (EmitsDecl)emitters[i];
            // find the operation declarations
            OperationDecl opconnect = null;
            OperationDecl opdisconnect = null;
            OperationDecl[] opeqs = emitter.getClientMapping();
            for (int j=0;j<opeqs.length;j++) {
                if (opeqs[j].getName().charAt(0)=='c') {
                    opconnect = opeqs[j];
                }
                else if (opeqs[j].getName().charAt(0)=='d') {
                    opdisconnect = opeqs[j];
                }
            }

            // * final public void
            //   connect_${emitter.getName()}(${emitter.getEvent().getName()}Consumer consumer)
            //   throws org.omg.Components.AlreadyConnected
            addJavaOperation(ski, opconnect, opconnect.getType(),
                             opconnect.getParameterList().getParameters(),
                             opconnect.getExceptionList().getExceptions(),
                             "MAINSEGBASE_EMITTER_CONNECT_OPERATION_BODY",
                             "emitter", emitter);

            // * final public ${emitter.getEvent().getName()}Consumer
            //   disconnect_${emitter.getName()}()
            //   throws org.omg.Components.NoConnection
            addJavaOperation(ski, opdisconnect, opdisconnect.getType(),
                             opdisconnect.getParameterList().getParameters(),
                             opdisconnect.getExceptionList().getExceptions(),
                             "MAINSEGBASE_EMITTER_DISCONNECT_OPERATION_BODY",
                             "emitter", emitter);
        }

        // publishers equiv operation
        for (int i=0;i<publishers.length;i++) {
            PublishesDecl publisher = (PublishesDecl)publishers[i];
            // find the operation declarations
            OperationDecl opsubscribe = null;
            OperationDecl opunsubscribe = null;
            OperationDecl[] opeqs = publisher.getClientMapping();
            for (int j=0;j<opeqs.length;j++) {
                if (opeqs[j].getName().charAt(0)=='s') {
                    opsubscribe = opeqs[j];
                }
                else if (opeqs[j].getName().charAt(0)=='u') {
                    opunsubscribe = opeqs[j];
                }
            }

            // * final public org.omg.Components.Cookie
            //   connect_${publisher.getName()}(${publisher.getEvent().getName()}Consumer consumer)
            //   throws org.omg.Components.ExceededConnectionLimit
            addJavaOperation(ski, opsubscribe, opsubscribe.getType(),
                             opsubscribe.getParameterList().getParameters(),
                             opsubscribe.getExceptionList().getExceptions(),
                             "MAINSEGBASE_PUBLISHER_SUBSCRIBE_OPERATION_BODY",
                             "publisher", publisher);

            // * final public ${publisher.getEvent().getName()}Consumer
            //   disconnect_${publisher.getName()}(org.omg.Components.Cookie ck)
            //   throws org.omg.Components.InvalidConnection
            addJavaOperation(ski, opunsubscribe, opunsubscribe.getType(),
                             opunsubscribe.getParameterList().getParameters(),
                             opunsubscribe.getExceptionList().getExceptions(),
                             "MAINSEGBASE_PUBLISHER_UNSUBSCRIBE_OPERATION_BODY",
                             "publisher", publisher);
        }

        // add to repository
        getRepository().addObject(ski);
    }

    private void
    addJavaOperation(ClassObject ski, Declaration decl, TypeRef rtype,
                     Parameter[] params, ExceptionDecl[] exceptions,
                     String macroname)
    {
        addJavaOperation(ski, decl, rtype, params, exceptions, macroname, new String[0], new Object[0]);
    }

    private void
    addJavaOperation(ClassObject ski, Declaration decl, TypeRef rtype,
                     Parameter[] params, ExceptionDecl[] exceptions,
                     String macroname, String ctxkey, Object ctxval)
    {
        String[] ctxkeys = { ctxkey };
        Object[] ctxvals = { ctxval };

        addJavaOperation(ski, decl, rtype, params, exceptions, macroname, ctxkeys, ctxvals);
    }

    private void
    addJavaOperation(ClassObject ski, Declaration decl, TypeRef rtype,
                     Parameter[] params, ExceptionDecl[] exceptions, String macroname,
                     String[] ctxkeys, Object[] ctxvals)
    {
        MethodObject operation = new MethodObjectImpl();
        // return type (java mapping)
        operation.setReturnType(getTranslator().toJava(rtype));
        
        // name
        operation.setName(decl.getName());
        
        // parameters (java mapping)
        ParameterObject parameter = null;
        for (int j=0;j<params.length;j++) {
            parameter = new ParameterObjectImpl();
            parameter.setName(params[j].getName());
            parameter.setType(getTranslator().toJava(params[j].getType(), params[j].getMode()));
            operation.addParameter(parameter);
        }
        // build parameters list
        String param_names = "";
        // check if params set is not empty
        if (params.length!=0) {
            for (int j=0;j<params.length-1;j++) {
                param_names = param_names + params[j].getName() + ", ";
            }
            
            param_names = param_names + params[params.length-1].getName();
        }
        operation.getImpl().addContextValue("parameter_list", param_names);
        
        // exceptions
        if (exceptions.length!=0) {
            operation.getImpl().addContextValue("has_exceptions", new Boolean(true));
            String[] exc_absnames = new String[exceptions.length];
            for (int j=0;j<exceptions.length;j++) {
                exc_absnames[j] = getTranslator().getAbsoluteName(exceptions[j]);
                    operation.addException(exc_absnames[j]);
            }
            operation.getImpl().addContextValue("exception_absnames", exc_absnames);
        }
        else {
                operation.getImpl().addContextValue("has_exceptions", new Boolean(false));
        }

        // body
        // NOTE: only the body's macro differs from a client-side to a server-side wrapped operation
        operation.getImpl().setMacro(macroname);
        operation.getImpl().addContextValue("declaration", decl);
        operation.getImpl().addContextValue("return_type", getTranslator().toJava(rtype));

        // add external context values
        for (int i=0;i<ctxkeys.length;i++) {
            operation.getImpl().addContextValue(ctxkeys[i], ctxvals[i]);
        }

        if (rtype.getTypeKind()!=TypeKind.tk_void) {
            operation.getImpl().addContextValue("has_return_type", new Boolean(true));
        }
        else {
            operation.getImpl().addContextValue("has_return_type", new Boolean(false));
        }

        // add operation to class
        ski.addMethod(operation);
    }

    //
    //  public operations
    //

    final public void
    generate(String basedir, Scope scope)
    throws GenerationException
    {
        _repository = new RepositoryImpl();

        generateForDeclaration(scope);

        // NOTE: issue: the template file "cidl_jimpl.vm" is not located in the OpenCCM template dir,
        // but in the directory pointed by the "ecm.templates" property
        // It could be usefull to be able to specify externally the template dirs
        // in the initialization of the Java generator (along with the template files
        // themselves)
        // Currently, we are forced to copy the template file in the OpenCCM dir ...

        String tpath = System.getProperty("ecm.templates");
        java.io.File tdir = new java.io.File(tpath);

        // Generate Java Files
        JavaGenerator javagen = new JavaGenerator();
        java.util.Vector v = new java.util.Vector();
        v.add("cidl_jimpl.vm");
        javagen.initialize("Component Implementation", v);
        javagen.generate(getRepository(), basedir);
    }

}


