// ====================================================================
// 
//   ECM: The Extensible Container Model
//   Copyright (C) 2004 THALES
//   Contact: openccm-ecm@objectweb.org
//
//   This library is free software; you can redistribute it and/or
//   modify it under the terms of the GNU Lesser General Public
//   License as published by the Free Software Foundation; either
//   version 2.1 of the License, or any later version.
//
//   This library is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//   Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public
//   License along with this library; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
//   USA
//
//   Initial developer(s): Mathieu Vadet.
//   Initial Funding: IST COACH European project (IST-2001-34445)
//                    http://www.ist-coach.org 
//
// ====================================================================



package org.objectweb.ccm.runtime;

import org.objectweb.corba.runtime.*;

/**
 **  <p>Default system factory implementation for the <tt>ContainerService</tt> system service.</p>
 **/
public class ContainerFactoryImpl
extends org.omg.CORBA.LocalObject
implements SystemFactory
{
    // system factory
    static final private String _class_name = "ContainerFactoryImpl";

    // default constructor
    protected
    ContainerFactoryImpl()
    {
    }

    //
    //  entry point
    //

    static public SystemFactory
    create_factory()
    {
        return new ContainerFactoryImpl();
    }

    //
    //  internal operations
    //

    static private ServiceImpl[]
    asServices(String[] uuids, ORBService orbs)
    {
        final String opname = "asServices";
        String msg = null;

        // check if uuids is empty
        if ((uuids==null) || (uuids.length==0)) {
            return new ServiceImpl[0];
        }

        // NOTE: use an array list rather than an array as some uuids may be skipped
        // (ie we can't foresee the size of the array)
        java.util.ArrayList services = new java.util.ArrayList();

        // get ExtComponentInstallation
        // later used for CCSD, CSD, and code obtention
        org.omg.CORBA.Object obj = null;
        org.coach.ECM.ExtComponentInstallation extci = null;

        obj = orbs.resolve_initial_references("ExtComponentInstallation");
        extci = org.coach.ECM.ExtComponentInstallationHelper.narrow(obj);

        ServiceImpl service = null;
        org.coach.ECM.ServiceArchive sarchive = null;
        for (int i=0;i<uuids.length;i++) {
            // obtain service archive
            sarchive = extci.get_service_archive(uuids[i]);

            // create service with:
            //    - ServiceId
            //    - Implementation UUID
            //    - reference to the service set
            String serviceid = sarchive.service_id();
            service = new ServiceImpl(serviceid, uuids[i]);

            // check if the service requires a corba service
            org.coach.ECM.CORBAServiceRef rcsref = null;

            try {
                rcsref = sarchive.get_corbaservice_ref();

                // create corba service ref
                CORBAServiceRef csref = new CORBAServiceRef();
                csref.stringified_initial_reference = orbs.object_to_string(rcsref.initial_reference);
                csref.corba_type = rcsref.corba_type;
                csref.orb_registration_name = rcsref.orb_registration_name;

                // store
                service.setCorbaServiceRef(csref);
            } catch (org.coach.ECM.NoCORBAServiceRefRequired ex) {
                // ignore
                TheLogger.debug(_class_name, opname, "IGNORE", ex);
            }

            // create service initializer
            String codeloc = sarchive.code_location();
            TheClassLoader.addResource(codeloc);

            msg = "Code location for service: "+uuids[i]+" is: "+codeloc;
            TheLogger.debug(_class_name, opname, msg);

            // create new instance
            org.coach.ECA.ServiceInitializer serviceinit = null;
            String entrypt = sarchive.code_entrypoint();

            msg = "Code entrypoint for service: "+uuids[i]+" is: "+entrypt;
            TheLogger.debug(_class_name, opname, msg);

            try {
                serviceinit = (org.coach.ECA.ServiceInitializer)TheClassLoader.newInstance(entrypt);
            } catch (ClassCastException ex) {
                // NOTE: TODO: how to handle this case ? pass the service installation ? this failure should
                // be reported or we consider that the case can not happen (because the archive was checked)
                msg = "FAILED (not a service initializer)";
                TheLogger.debug(_class_name, opname, msg);
                // skip
                continue ;
            }

            // set 
            service.setServiceInitializer(serviceinit);

            // store
            services.add(service);
        }

        //
        return (ServiceImpl[])services.toArray(new ServiceImpl[0]);
    }

    static private ServiceImpl[]
    asServices(String[] uuids, String[] sids, String[] arch_locs, String[] entrypts,
               String[] iref_regnames, String[] iref_iors)
    {
        final String opname = "asServices";
        String msg = null;

        // NOTE: this operation should be used for the first ORB
        // indeed, as no ORB is available, it is not possible to obtain the
        // ExtComponentInstallation ref to query the information

        // check if uuids is empty
        if ((uuids==null) || (uuids.length==0)) {
            return new ServiceImpl[0];
        }

        // NOTE: use an array list rather than an array as some uuids may be skipped
        // (ie we can't foresee the size of the array)
        java.util.ArrayList services = new java.util.ArrayList();
        ServiceImpl service = null;

        for (int i=0;i<uuids.length;i++) {
            service = new ServiceImpl(sids[i], uuids[i]);

            // check if service has an initial reference
            if ((iref_regnames[i]!=null) && (!iref_regnames.equals(""))) {
                // check that IOR is set
                if ((iref_iors[i]==null) || (iref_iors[i].equals(""))) {
                    msg = "FAILED (no IOR for: "+iref_regnames[i]+")";
                    TheLogger.debug(_class_name, opname, msg);
                    // skip
                    continue ;
                } 

                // create CORBA service ref
                CORBAServiceRef csref = new CORBAServiceRef();
                csref.corba_type = "unknown";
                csref.orb_registration_name = iref_regnames[i];
                csref.stringified_initial_reference = iref_iors[i];
                csref.initial_reference = null;

                // store
                service.setCorbaServiceRef(csref);
            }

            // create service initializer
            String codeloc = arch_locs[i];
            TheClassLoader.addResource(codeloc);

            // create new instance
            org.coach.ECA.ServiceInitializer serviceinit = null;
            String entrypt = entrypts[i];

            try {
                serviceinit = (org.coach.ECA.ServiceInitializer)TheClassLoader.newInstance(entrypt);
            } catch (ClassCastException ex) {
                // NOTE: TODO: how to handle this case ? pass the service installation ? this failure should
                // be reported or we consider that the case can not happen (because the archive was checked)
                msg = "FAILED (not a service initializer)";
                TheLogger.debug(_class_name, opname, msg);
                // skip this uuid
                continue ;
            }

            // set 
            service.setServiceInitializer(serviceinit);

            // store
            services.add(service);
        }

        //
        return (ServiceImpl[])services.toArray(new ServiceImpl[0]);
    }

    static private CORBAServiceRef[]
    asDegenServices(org.coach.ECM.DegeneratedServiceRef[] dsrefs,
                    ORBService orbs)
    {
        // check if desrefs is empty
        if ((dsrefs==null) || (dsrefs.length==0)) {
            return new CORBAServiceRef[0];
        }

        CORBAServiceRef[] dservices = new CORBAServiceRef[dsrefs.length];

        for (int i=0;i<dsrefs.length;i++) {
            // create CORBA Service ref
            dservices[i] = new CORBAServiceRef();
            dservices[i].corba_type = dsrefs[i].corba_type;
            dservices[i].orb_registration_name = dsrefs[i].orb_registration_name;
            dservices[i].stringified_initial_reference = dsrefs[i].initial_reference;
            dservices[i].initial_reference = orbs.string_to_object(dsrefs[i].initial_reference);
        }

        return dservices;
    }

    static private CORBAServiceRef[]
    asDegenServices(String[] iref_regnames, String[] iref_iors)
    {
        // NOTE: this operation should be used for the first ORB
        // indeed, as no ORB is available, it is not possible to obtain the
        // ExtComponentInstallation ref to query the information

        // check if empty
        if ((iref_regnames==null) || (iref_regnames.length==0)) {
            return new CORBAServiceRef[0];
        }

        // NOTE: use an array list rather than an array as some uuids may be skipped
        // (ie we can't foresee the size of the array)
        java.util.ArrayList dservices = new java.util.ArrayList();
        CORBAServiceRef dservice = null;

        for (int i=0;i<iref_regnames.length;i++) {
            // check that IOR is set
            if ((iref_iors[i]==null) || (iref_iors[i].equals(""))) {
                final String opname = "asDegenServices";
                final String msg = "IGNORE (no IOR for: "+iref_regnames[i]+")";
                TheLogger.debug(_class_name, opname, msg);
                // skip
                continue ;
            } 

            // create CORBA Service ref
            dservice = new CORBAServiceRef();
            dservice.corba_type = "unknown";
            dservice.orb_registration_name = iref_regnames[i];
            dservice.stringified_initial_reference = iref_iors[i];
            dservice.initial_reference = null;
            dservices.add(dservice);
        }

        return (CORBAServiceRef[])dservices.toArray(new CORBAServiceRef[0]);
    }

    static private ORBService
    createORBService(ServiceImpl[] services,
                     CORBAServiceRef[] dservices)
    {
        // create service init info
        ServiceInitInfoImpl sii = new ServiceInitInfoImpl(services, dservices);

        // run pre_install 
        org.coach.ECA.ServiceInitializer sinit = null;
        for (int i=0;i<services.length;i++) {
            sinit = services[i].getServiceInitializer();
            sinit.pre_install(sii);
        }

        // obtain FactoryFinder service
        FactoryFinderService ffservice = ComponentRuntime.getFFService();

        // obtain factories
        SystemFactory orbfact = ffservice.find_service_factory(ORBService.SERVICE_ID);

        // build init info
        ORBConfiguration orbcfg = sii.getORBConfiguration();

        // create ORBService
        ORBService orbservice = (ORBService)orbfact.create_system_component(orbcfg);
        // set ORB service for post_install
        sii.setORBService(orbservice);

        // run post_install
        for (int i=0;i<services.length;i++) {
            sinit = services[i].getServiceInitializer();
            sinit.post_install(sii);
        }

        //
        return orbservice;
    }

    // NOTE: no args currently, but it will change
    static private ComponentPOAService
    createComponentPOAService(String poasid,
                              ORBService orbs,
                              ServicesSet sset)
    {
        // obtain FactoryFinder service
        FactoryFinderService ffservice = ComponentRuntime.getFFService();

        // obtain factories
        SystemFactory poafact = ffservice.find_service_factory(poasid);

        // build init info
        ComponentPOAConfiguration poacfg = new ComponentPOAConfigurationImpl(null, null, orbs, sset);

        // create service
        ComponentPOAService poas = (ComponentPOAService)poafact.create_system_component(poacfg);

        //
        return poas;
    }

    //
    //  public operations
    //

    // create a container service with:
    //   - an already created component installation registered to the ORB service
    //   - a set of services uuids
    //   - a set of degenerated service refs
    // NOTE: this is the default op
    static public ContainerService
    asContainerService(String[] uuids,
                       org.coach.ECM.DegeneratedServiceRef[] dsrefs,
                       ORBService orbs,
                       String poasid)
    {
        ServiceImpl[] services = null;
        CORBAServiceRef[] dservices = null;
        ORBService orbservice = null;
        ServicesSetImpl sset = null;
        ComponentPOAService cpoaservice = null;
        
        // create services
        services = asServices(uuids, orbs);

        // create degenerated services
        dservices = asDegenServices(dsrefs, orbs);

        // create ORB service
        // NOTE: the ORB service is created in this class as at this time
        // the ServicesSetImpl class is mainly a storage class
        orbservice = createORBService(services, dservices);

        // create services set
        sset = new ServicesSetImpl(services, dservices, orbservice);

        // create component poa service component
        cpoaservice = createComponentPOAService(poasid, orbservice, sset);

        // create container service
        ContainerService contservice = new ContainerServiceImpl(orbservice, cpoaservice, sset);

        return contservice;
    }

    // create a container service with:
    //   - no ORB service
    //   - a set of services uuids
    //   - a set of degenerated service refs
    //   - the ExtendedComponentInstallation ref
    // NOTE: this will create a "empty" ORB service for remote comms between this runtime
    // and the ExtCI
    static public ContainerService
    asContainerService(String[] uuids,
                       org.coach.ECM.DegeneratedServiceRef[] dsrefs,
                       StringifiedInitialReference[] sirefs,
                       String poasid)
    {
        // create an "empty" ORB service with the ExtCI ref as initial reference
        ORBService orbservice = null;

        // obtain FactoryFinder service
        FactoryFinderService ffservice = ComponentRuntime.getFFService();

        // obtain factories
        SystemFactory orbfact = ffservice.find_service_factory(ORBService.SERVICE_ID);

        // build config
        // add stringified initial references only
        ORBConfiguration orbcfg = new ORBConfigurationImpl(null, null, null, sirefs);

        // create ORBService
        orbservice = (ORBService)orbfact.create_system_component(orbcfg);

        // use the default op with this ORB service
        return asContainerService(uuids, dsrefs, orbservice, poasid);
    }

    // create a services set with:
    //   - no ORB service
    //   - no ExtCI ref (ie no ExtCI is installed on the node)
    //   - a set of services uuids and archive locations
    //   - a set of degenerated service refs
    // NOTE: this should create the ExtCI
    // TODO

    //
    //  IDL:objectweb.org/corba/runtime/SystemFactory:1.0
    //

    final public SystemComponent
    create_system_component(SystemConfiguration cfg)
    {
        final String opname = "create_system_component";
        final String msg = "operation not implemented";
        TheLogger.debug(_class_name, opname, msg);
        throw new Error(msg);
    }
}
